==> intel-oneapi-mkl-2022.1.0_223-1-aarch64.pkg.tar.xz: f_offload/dft/source/sp_complex_1d_batch.f90 <==
!===============================================================================
! Copyright 2021-2022 Intel Corporation.
!
! This software and the related documents are Intel copyrighted  materials,  and
! your use of  them is  governed by the  express license  under which  they were
! provided to you (License).  Unless the License provides otherwise, you may not
! use, modify, copy, publish, distribute,  disclose or transmit this software or
! the related documents without Intel's prior written permission.
!
! This software and the related documents  are provided as  is,  with no express
! or implied  warranties,  other  than those  that are  expressly stated  in the
! License.
!===============================================================================

! Content:
! A simple example of batch single-precision complex-to-complex in-place 1D
! FFTs using Intel(R) oneAPI Math Kernel Library (oneMKL) DFTI
!
!*****************************************************************************

include "mkl_dfti_omp_offload.f90"

program sp_complex_1d_batch

  use MKL_DFTI_OMP_OFFLOAD, forget => DFTI_SINGLE, DFTI_SINGLE => DFTI_SINGLE_R
  use omp_lib, ONLY : omp_get_num_devices
  use, intrinsic :: ISO_C_BINDING
  ! Size of 1D transform
  integer, parameter :: N = 16
  
  ! Number of transforms
  integer, parameter :: M = 7
  
  integer, parameter :: numPoints = M*N;

  ! Arbitrary harmonic used to verify FFT
  integer, parameter :: H = 1

  ! Working precision is single precision
  integer, parameter :: WP = selected_real_kind(6,37)

  ! Execution status
  integer :: status = 0, ignored_status

  ! The data array
  complex(WP), allocatable :: x (:)

  ! DFTI descriptor handle
  type(DFTI_DESCRIPTOR), POINTER :: hand

  hand => null()

  print *,"Example sp_complex_1d_batch"
  print *,"Batch forward and backward single-precision ",        &
    &      "complex-to-complex in-place 1D transform"
  print *,"Configuration parameters:"
  print *,"DFTI_PRECISION      = DFTI_SINGLE"
  print *,"DFTI_FORWARD_DOMAIN = DFTI_COMPLEX"
  print *,"DFTI_DIMENSION      = 1"
  print '(" DFTI_NUMBER_OF_TRANSFORMS = "I0"")', M
  print '(" DFTI_LENGTHS        = /"I0"/" )', N

  print *,"Create DFTI descriptor"
  status = DftiCreateDescriptor(hand, DFTI_SINGLE, DFTI_COMPLEX, 1, N)
  if (0 /= status) goto 999
  
  print *,"Set DFTI descriptor for number of batch in-place transforms"
  status = DftiSetValue(hand, DFTI_NUMBER_OF_TRANSFORMS, M)
  if (0 /= status) goto 999
  
  print *,"Set DFTI descriptor for size of individual transforms"
  status = DftiSetValue(hand, DFTI_INPUT_DISTANCE, N)
  if (0 /= status) goto 999

  print *,"Commit DFTI descriptor"
  !$omp target variant dispatch
  status = DftiCommitDescriptor(hand)
  !$omp end target variant dispatch
  if (0 /= status) goto 999

  print *,"Allocate array for input data"
  allocate ( x(numPoints), STAT = status)
  if (0 /= status) goto 999

  print *,"Initialize input for forward transforms"
  call init(x, M, N, H)

  print *,"Compute forward transforms"
  !$omp target data map(tofrom:x)
  !$omp target variant dispatch use_device_ptr(x)
  status = DftiComputeForward(hand, x)
  !$omp end target variant dispatch
  !$omp end target data
  if (0 /= status) goto 999

  print *,"Verify the results"
  status = verify(x, M, N, H)
  if (0 /= status) goto 999

  print *,"Initialize input for backward transforms"
  call init(x, M, N, -H)

  print *,"Compute backward transforms"
  !$omp target data map(tofrom:x)
  !$omp target variant dispatch use_device_ptr(x)
  status = DftiComputeBackward(hand, x)
  !$omp end target variant dispatch
  !$omp end target data
  if (0 /= status) goto 999

  print *,"Verify the results"
  status = verify(x, M, N, H)
  if (0 /= status) goto 999

100 continue

  print *,"Release the DFTI descriptor"
  ignored_status = DftiFreeDescriptor(hand)

  if (allocated(x)) then
      print *,"Deallocate data array"
      deallocate(x)
  endif

  if (status == 0) then
    print *,"TEST PASSED"
    call exit(0)
  else
    print *,"TEST FAILED"
    call exit(1)
  endif

999 print '("  Error, status = ",I0)', status
  goto 100

contains

  ! Compute mod(K*L,M) accurately
  pure real(WP) function moda(k,l,m)
    integer, intent(in) :: k,l,m
    integer*8 :: k8
    k8 = k
    moda = real(mod(k8*l,m),WP)
  end function moda

  ! Initialize array with harmonic H
  subroutine init(x, M, N, H)
    integer M, N, H
    complex(WP) :: x(:)

    integer j, k
    complex(WP), parameter :: I_TWOPI = (0.0_WP,6.2831853071795864769_WP)

  do j = 1, M
    do k = 1, N
      x((j-1)*N + k) = exp( I_TWOPI * moda(k-1, H, N) / cmplx(N) ) / cmplx(N)
    end do
  end do
  end subroutine init

  ! Verify that x(N) is unit peak at x(H)
  integer function verify(x, M, N, H)
    integer M, N, H
    complex(WP) :: x(:)

    integer j, k
    real(WP) err, errthr, maxerr
    complex(WP) res_exp, res_got

    ! Note, this simple error bound doesn't take into account error of
    ! input data
    errthr = 5.0 * log(real(N, WP)) / log(2.0_WP) * EPSILON(1.0_WP)
    print '("  Check if err is below errthr " G10.3)', errthr

    maxerr = 0.0_WP
  do j = 1, M
      do k = 1, N
      if (mod(k-1-H,N)==0) then
      res_exp = 1.0_WP
      else
      res_exp = 0.0_WP
      end if
      res_got = x((j-1)*N + k)
      err = abs(res_got - res_exp)
      maxerr = max(err,maxerr)
      if (.not.(err < errthr)) then
      print '("  x("I0"): "$)', (j-1)*N + k
      print '(" expected ("G14.7", "G14.7"),"$)', res_exp
      print '(" got ("G14.7", "G14.7"),"$)', res_got
      print '(" err "G10.3)', err
      print *," Verification FAILED"
      verify = 100
      return
      end if
    end do
  end do
    print '("  Verified,  maximum error was " G10.3)', maxerr
    verify = 0
  end function verify

end program sp_complex_1d_batch

==> intel-oneapi-mkl-2023.2.0_49495-2-aarch64.pkg.tar.xz: f_offload/dft/source/sp_complex_1d_batch.f90 <==
!===============================================================================
! Copyright 2021-2022 Intel Corporation.
!
! This software and the related documents are Intel copyrighted  materials,  and
! your use of  them is  governed by the  express license  under which  they were
! provided to you (License).  Unless the License provides otherwise, you may not
! use, modify, copy, publish, distribute,  disclose or transmit this software or
! the related documents without Intel's prior written permission.
!
! This software and the related documents  are provided as  is,  with no express
! or implied  warranties,  other  than those  that are  expressly stated  in the
! License.
!===============================================================================

! Content:
! A simple example of batch single-precision complex-to-complex in-place 1D
! FFTs using Intel(R) oneAPI Math Kernel Library (oneMKL) DFTI
!
!*****************************************************************************

include "mkl_dfti_omp_offload.f90"

program sp_complex_1d_batch

  use MKL_DFTI_OMP_OFFLOAD, forget => DFTI_SINGLE, DFTI_SINGLE => DFTI_SINGLE_R
  use omp_lib, ONLY : omp_get_num_devices
  use, intrinsic :: ISO_C_BINDING
  ! Size of 1D transform
  integer, parameter :: N = 16
  
  ! Number of transforms
  integer, parameter :: M = 7
  
  integer, parameter :: numPoints = M*N;

  ! Arbitrary harmonic used to verify FFT
  integer, parameter :: H = 1

  ! Working precision is single precision
  integer, parameter :: WP = selected_real_kind(6,37)

  ! Execution status
  integer :: status = 0, ignored_status

  ! The data array
  complex(WP), allocatable :: x (:)

  ! DFTI descriptor handle
  type(DFTI_DESCRIPTOR), POINTER :: hand

  hand => null()

  print *,"Example sp_complex_1d_batch"
  print *,"Batch forward and backward single-precision ",        &
    &      "complex-to-complex in-place 1D transform"
  print *,"Configuration parameters:"
  print *,"DFTI_PRECISION      = DFTI_SINGLE"
  print *,"DFTI_FORWARD_DOMAIN = DFTI_COMPLEX"
  print *,"DFTI_DIMENSION      = 1"
  print '(" DFTI_NUMBER_OF_TRANSFORMS = "I0"")', M
  print '(" DFTI_LENGTHS        = /"I0"/" )', N

  print *,"Create DFTI descriptor"
  status = DftiCreateDescriptor(hand, DFTI_SINGLE, DFTI_COMPLEX, 1, N)
  if (0 /= status) goto 999
  
  print *,"Set DFTI descriptor for number of batch in-place transforms"
  status = DftiSetValue(hand, DFTI_NUMBER_OF_TRANSFORMS, M)
  if (0 /= status) goto 999
  
  print *,"Set DFTI descriptor for size of individual transforms"
  status = DftiSetValue(hand, DFTI_INPUT_DISTANCE, N)
  if (0 /= status) goto 999

  print *,"Commit DFTI descriptor"
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch
#else
  !$omp target variant dispatch
#endif
  status = DftiCommitDescriptor(hand)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  if (0 /= status) goto 999

  print *,"Allocate array for input data"
  allocate ( x(numPoints), STAT = status)
  if (0 /= status) goto 999

  print *,"Initialize input for forward transforms"
  call init(x, M, N, H)

  print *,"Compute forward transforms"
  !$omp target data map(tofrom:x)
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch
#else
  !$omp target variant dispatch use_device_ptr(x)
#endif
  status = DftiComputeForward(hand, x)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  !$omp end target data
  if (0 /= status) goto 999

  print *,"Verify the results"
  status = verify(x, M, N, H)
  if (0 /= status) goto 999

  print *,"Initialize input for backward transforms"
  call init(x, M, N, -H)

  print *,"Compute backward transforms"
  !$omp target data map(tofrom:x)
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch
#else
  !$omp target variant dispatch use_device_ptr(x)
#endif
  status = DftiComputeBackward(hand, x)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  !$omp end target data
  if (0 /= status) goto 999

  print *,"Verify the results"
  status = verify(x, M, N, H)
  if (0 /= status) goto 999

100 continue

  print *,"Release the DFTI descriptor"
  ignored_status = DftiFreeDescriptor(hand)

  if (allocated(x)) then
      print *,"Deallocate data array"
      deallocate(x)
  endif

  if (status == 0) then
    print *,"TEST PASSED"
    call exit(0)
  else
    print *,"TEST FAILED"
    call exit(1)
  endif

999 print '("  Error, status = ",I0)', status
  goto 100

contains

  ! Compute mod(K*L,M) accurately
  pure real(WP) function moda(k,l,m)
    integer, intent(in) :: k,l,m
    integer*8 :: k8
    k8 = k
    moda = real(mod(k8*l,m),WP)
  end function moda

  ! Initialize array with harmonic H
  subroutine init(x, M, N, H)
    integer M, N, H
    complex(WP) :: x(:)

    integer j, k
    complex(WP), parameter :: I_TWOPI = (0.0_WP,6.2831853071795864769_WP)

  do j = 1, M
    do k = 1, N
      x((j-1)*N + k) = exp( I_TWOPI * moda(k-1, H, N) / cmplx(N) ) / cmplx(N)
    end do
  end do
  end subroutine init

  ! Verify that x(N) is unit peak at x(H)
  integer function verify(x, M, N, H)
    integer M, N, H
    complex(WP) :: x(:)

    integer j, k
    real(WP) err, errthr, maxerr
    complex(WP) res_exp, res_got

    ! Note, this simple error bound doesn't take into account error of
    ! input data
    errthr = 5.0 * log(real(N, WP)) / log(2.0_WP) * EPSILON(1.0_WP)
    print '("  Check if err is below errthr " G10.3)', errthr

    maxerr = 0.0_WP
  do j = 1, M
      do k = 1, N
      if (mod(k-1-H,N)==0) then
      res_exp = 1.0_WP
      else
      res_exp = 0.0_WP
      end if
      res_got = x((j-1)*N + k)
      err = abs(res_got - res_exp)
      maxerr = max(err,maxerr)
      if (.not.(err < errthr)) then
      print '("  x("I0"): "$)', (j-1)*N + k
      print '(" expected ("G14.7", "G14.7"),"$)', res_exp
      print '(" got ("G14.7", "G14.7"),"$)', res_got
      print '(" err "G10.3)', err
      print *," Verification FAILED"
      verify = 100
      return
      end if
    end do
  end do
    print '("  Verified,  maximum error was " G10.3)', maxerr
    verify = 0
  end function verify

end program sp_complex_1d_batch
